import socket
from blinkenbase.Color import Color

STANDARD_WIDTH = 3.5

COLOR_BLACK = Color((0, 0, 0))

class Blinkenfoo(object):
    """This class represents a blinkenfoo instance.
    A blinkenfoo can display information send via upd and can listen to mqtt to play
    preconfigured animations. It als knows how many pixels it has:
    A pixel is one blinken-unit which can be controlled separately from the other
    units. A pixel can be a single LED on a i.E. ws2812 or a whole stripe.
    If the blinkenfoo is a led matrix then the pixel count is a*b.
    The physical width describes the width in cm of one blinkenunit/pixel. This is
    to make animations run the same speed on different blinkenfoo devices"""

    def __init__(self, p_name, p_udp_host, p_length, p_physikal_width = STANDARD_WIDTH, p_udp_port = 2390, p_brightness = 100):
        self.name = p_name
        self.udp_host = p_udp_host
        self.udp_port = p_udp_port

        self.length = p_length
        self.physikal_width = p_physikal_width
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.buffer = None
        self._reset_buffer()

        self.current_fps = None
        self.brightness_factor = p_brightness

    def __setattr__(self, p_key, p_value):
        if p_key == "brightness_factor":
            f = 0
            if p_value <= 100:
                f = 0.01 * p_value
            else:
                f = 100.0
            p_value = f
        super(Blinkenfoo, self).__setattr__(p_key, p_value)

    def _split_to_byte(self, p_number):
        """
        Splits an integer into an array of two bytes. the lower bytes are in [1]
        """
        return [(p_number >> 8) & 0xFF, p_number & 0xFF]
        #return [0, p_number]

    def _generate_package(self, startpixel, endpixel, rgb):
            """ Generates a udp package the blinkenfoo understands"""
            pack = bytearray()
            startpixel = self._split_to_byte(startpixel)
            endpixel = self._split_to_byte(endpixel)
            pack.extend(startpixel)
            pack.extend(endpixel)
            for color in rgb:
                r, g, b = color
                pack.extend([r, g, b])
            return pack

    def _send_package(self, p_package):
        """Sends the given udp_package to this's host:port"""
        try:
            self.udp_socket.sendto(p_package, (self.udp_host, self.udp_port))
        except socket.error:
            pass

    def _reset_buffer(self):
        """
        Sets the content of the buffer to black.
        """
        self.buffer = [COLOR_BLACK for _ in range(self.length)]

    ####################
    ###Public methods###
    ####################

    def is_reachable(self):
        """
        Checks wether a blinkenfoo is reachable or not. sends a dummy package.
        """
        rgb_frame = [(0, 0, 0) for _ in range(self.length)]
        package = self._generate_package(0, 1, [(0, 0, 0)])
        try:
            self.udp_socket.sendto(package, (self.udp_host, self.udp_port))
            return True
        except socket.error:
            return False

    def calc_fps(self, p_fps):
        """
        Calculates fps according to this blinkenfoo width in order for an
        animation to run the same "speed" on every device.
        """
        return STANDARD_WIDTH * p_fps / self.physikal_width

    def send_frame(self, p_frame):
        """
        Sends an frame to the blinkenfoo. A frame is a list of color objects.
        Other types may cause erros later in the programm. The length of the
        frame must match the pixel count of this blinkenfoo.
        """
        length = len(p_frame)
        if length == self.length:
            rgb_frame = []
            self.buffer = p_frame
            for i in range(len(p_frame)):
                r, g, b = p_frame[i].rgb_tuple(True)
                r = int(r * self.brightness_factor)
                g = int(g * self.brightness_factor)
                b = int(b * self.brightness_factor)
                rgb_frame.append((r, g, b))
            package = self._generate_package(0, length-1, rgb_frame)
        else:
            raise ValueError("Given list length does not match pixel count!: Given %d, expected %d" % (length, self.length))
        self._send_package(package)

    def set_full_color(self, p_color):
        """Displays the given color on every pixel of the blinkenfoo"""
        rgb_list = [p_color for _ in range(self.length)]
        self.send_frame(rgb_list)

    def set_pixel(self, p_index, p_color):
        """Sets an specific color at the given index. The change is visible after
        flush is called"""
        if p_index > self.length:
            raise ValueError("Out of range: given %d, length %d" % (p_index, self.length))
        self.buffer[p_index] = p_color

    def flush(self, p_reset=False):
        """
        Sends the content of the buffer to the blinkenfoo. If p_reset = True
        then the buffer is resetet (see _reset_buffer())
        """
        self.send_frame(self.buffer)
        if p_reset:
            self._reset_buffer()

    def clear(self):
        self._reset_buffer()
        self.flush()

##################
###End of class###
##################


SPHERES = Blinkenfoo("Spheres", "ESP_35D447.warpzone", 9, 20)
PANEL = Blinkenfoo("Panel", "ESP_35d9E4.warpzone", 8, 12)
WARP_SIGN = Blinkenfoo("Warp-Sign", "ESP_133C4C.warpzone", 1, 25)
DMX = Blinkenfoo("DMX", "10.0.3.27", 5, 10, p_brightness = 70)
CUBES = Blinkenfoo("Cubes", "cubes.warpzone", 8, 5)
TISCH = Blinkenfoo("Tisch", "tisch.warpzone", 700)

DEVICE_LIST_LOUNGE = [WARP_SIGN, SPHERES, PANEL, CUBES, DMX]
#DEVICE_LIST_LOUNGE = [PANEL]
#DEVICE_LIST_LOUNGE = [TISCH]
#DEVICE_LIST_LOUNGE = [SPHERES]