class Command(object):

    # TODO aliases

    def __init__(self, p_name=None, p_handler=None, p_usage="", p_help=""):
        if not p_name:
            raise ValueError("Command-name cannot be None")

        self.name = p_name.lower()
        self.handler = p_handler
        self.help = p_help
        self.usage = p_usage
        self._shell = None

    def print_usage(self):
        """Prints the usage of this command in the currently assigned shell"""
        if self._shell is not None:
            self._shell.print_notification(self.usage)
        else:
            raise AttributeError(
                "Command ''%s' is not assigned to a shell" % self.name
            )

    def assign_command_to_shell(self, p_shell):
        """
        Assigns this commmand to a shell. An command instance can only
        be assigned to one shell at a time
        """
        if not self._shell:
            self._shell = p_shell
        else:
            raise AttributeError(
                "Command '%s' is alread assigned to a shell" % self.name
            )

    def deassign_command_from_shell(self):
        """Signature says it all"""
        self._shell = None

    def get_shell(self):
        return self._shell


class Shell(object):

    def __init__(self, p_name="Programmer 2 lazy to name this shell"):
        self.SHELL_NAME = p_name
        self.SHELL_PREFIX = self.SHELL_NAME.lower() + "~: "

        self._RUNNING = False
        self._command_dict = {}

        command_exit = Command(
            p_name="exit",
            p_handler=self._exit_handler,
            p_usage="exit",
            p_help="Quits this shell"
        )
        command_help = Command(
            p_name="help",
            p_handler=self._help_handler,
            p_usage="help",
            p_help="Prints this help"
        )

        self.register_command(command_exit)
        self.register_command(command_help)

        # If no matching command was found, the default_handler is called
        self.default_handler = command_help.handler

    def evaluate_input(self, p_input):
        """
        Handles the given input as command.
        The deal is that a shell may not be always busy waiting, therefor
        a loop is not always necassery
        """
        # Splitting input string an removing whitespaces
        args = p_input.lower().split(" ")
        args = [a for a in args if a not in [" ", "", "\t"] ]

        # If no command was given
        if not len(args):
            self.default_handler(None, None)
            return

        command = self._get_command_by_name(args[0])

        # If given command was not found
        if command is None:
            self.print_error("Unknown command!")
            return
        #try:
        command.handler(command, args)
       # except Exception as e:
        #    self.print_error("\nThere was an error executing this command. Contact your admin/dev.\n" + str(e))

    # Shell main-control
    # ##################
    def start(self, p_busy_waiting=True):
        """
        Starting shell
        """
        self._RUNNING = True
        self.print_notification(
            "Welcome to " + self.SHELL_NAME + ". Type help for help."
        )

        if not p_busy_waiting:
            return

        while self._RUNNING:
            # Reading command
            input = self.read_command()
            self.evaluate_input(input)

    def stop(self):
        """
        Stopping shell
        """
        self.print_notification("Quitting " + self.SHELL_NAME + "...")
        self._RUNNING = False

    # Reading stuff from UI
    #######################
    def read_command(self):
        pass

    def read_string(self, p_message, p_default=None, p_retries=1):
        """
        Prints given message and reads user input as string.
        Prints error if input is not string/not decodeable and returns None.
        """
        return self._read_custom_value(
            p_message=p_message,
            p_cast_function=lambda x: x,
            p_retries=p_retries,
            p_error_message="Input must be text!",
            p_default=p_default
        )

    def read_float(self, p_message, p_default=0, p_retries=1):
        """
        Prints given message and reads user input as float.
        Prints error if input is no float and returns None.
        """
        return self._read_custom_value(
            p_message=p_message,
            p_cast_function=lambda x: float(x),
            p_retries=p_retries,
            p_error_message="Input must be float!",
            p_default=p_default
        )

    def read_int(self, p_message, p_default=0, p_retries=1):
        """
        Prints given message and reads user input as int.
        Prints error if input is no int and returns None.
        """
        return self._read_custom_value(
            p_message=p_message,
            p_cast_function=lambda x: int(x),
            p_retries=p_retries,
            p_error_message="Input must be Integer!",
            p_default=p_default
        )

    def _read_custom_value(self, p_message, p_cast_function, p_retries, p_error_message, p_default):
        """
        Reads a custom value and tries to cast it. If casting fails, the given error-message is printed.
        if retries > 0, then the user is prompted to enter the value again till all retries were used.
        """
        read = self.read_custom_input(p_message)

        if not len(read):
            return p_default

        try:
            read = p_cast_function(read)
        except ValueError:
            self.print_error(p_error_message)
            if p_retries > 0:
                return self._read_custom_value(p_message, p_cast_function, p_retries - 1, p_error_message, p_default)
            return p_default
        return read

    def read_custom_input(self, p_message):
        """
        Reads input with custom
        """
        pass

    # Handler for integrated commands
    #################################
    def _exit_handler(self, p_command, p_args):
        self.stop()

    def _help_handler(self, p_command, p_args):
        self.print_help()

    # Printmethods
    # ############
    def print_message(self, p_string):
        pass

    def print_notification(self, p_string):
        pass

    def print_error(self, p_string):
        pass

    def print_warning(self, p_string):
        pass

    def print_help(self):
        """
        Remove since this shell works with mqtt input.
        Help is handled by client
        """
        pass

     # Command registering
    # ###################
    def register_command(self, p_command):
        """Registers an command to this shell. A command is identified by its name."""
        name = p_command.name
        if name in self._command_dict:
            raise ValueError("Command %s is already registered! Unregister first to replace" % str(p_command.name))
        self._command_dict[name] = p_command
        p_command.assign_command_to_shell(self)

    def unregister_command(self, p_name):
        """Unregisters an command. If there is no such name, nothing will happen"""
        if p_name in self._command_dict:
            command = self._command_dict[p_name]
            command.deassign_command_from_shell()
            del self._command_dict[p_name]
        else:
            raise ValueError("No command by the alias %s registered" % str(p_name))

    def _get_command_by_name(self, p_name):
        """Maps an alias to the command key. Returns None if there is no such alias"""
        if p_name is None or p_name not in self._command_dict:
            return None
        return self._command_dict[p_name]

    # I dont think this is needed here but I'll let it live...just in case
    # More lines -> more money...i guess
    # (This project is opensource)
    # SHUT UP!
    def merge_values(self, p_value_old, p_value_new):
        """
        Merges two values. If one of both is None, the other one is returned.
        If both values are not the same, the user is prompted to choose
        """
        # If both values are the same
        if p_value_old == p_value_new:
            return p_value_old

        # If no value is set
        if not p_value_old and not p_value_new:
            return None

        # If both values are set, user is prompted to decide which one to use
        if p_value_old is not None and p_value_new is not None:
            choice = self.read_custom_input(
                "Entry is already known. Do want to replace '%s' with '%s'? [Y]es/[N]o/[A]bort: " %
                (str(p_value_old), str(p_value_new))
            )
            choice = choice.lower()
            if choice in ["yes", "y"]:
                return p_value_new
            if choice in ["no", "n"]:
                return p_value_old
            self.print_message("Aborting...")
            return None

        # If only new value is set
        if p_value_old is None:
            return p_value_new

        # If only old value is set
        if p_value_new is None:
            return p_value_old
